#!/usr/bin/env python
"""
Axe Interpreter

An interpreter for programs written in the Axe Parser language for computers.

Author: Michael Lee (<michael.lee.0x2a@gmail.com>)
License: GNU Lesser GPL
"""

## System Modules ##
import operator

## My Modules ##
import io
import axelex
import axeparse
import meta
from meta import ExitException

_debug = False
_draw = False



class Interpreter():
    def __init__(self):
        self.vars = {'A':0, 'B':0, 'C':0, 'D':0, 'E':0,
                     'F':0, 'G':0, 'H':0, 'I':0, 'J':0,
                     'K':0, 'L':0, 'M':0, 'N':0, 'O':0, 
                     'P':0, 'Q':0, 'R':0, 'S':0, 'T':0, 
                     'U':0, 'V':0, 'W':0, 'X':0, 'Y':0, 
                     'Z':0, 'Theta':0, 'Ans':0}
        self.operators = ('+', '-', '*', '/', '<=', '>=', '<', '>', '=')
        
        self.emulator = io.Emulator()
        self._DRAW_CLRDRAW([])
        self._DRAW_DISPGRAPH([])
        
        print "\nInterpreter Ready.\n"
        return
    
    def run(self, ast):
        
        ast = self.prune(ast)
        if ast == None:
            return
        ast = ast[1]
        output = self.execute(ast)
        if _draw:
            self.execute(['DRAW', 'DISPGRAPH'])
        return output
    
    def execute(self, ast):
        if _debug:
            print "execute:", ast
        
        self.emulator.sanity_check()
        
        if isinstance(ast, int):
            # Int check.
            return ast
        elif isinstance(ast, basestring):
            # String check
            if ast.startswith('VAR_'):
                return self.vars[ast[-1]]
        elif isinstance(ast, list):
            if len(ast) == 1:
                self.execute(ast[0])
            else:
                # List check.
                func = ast[0]
                args = ast[1:]
                
                value = getattr(self, '_' + func)(args)
                if _debug:
                    print "...returned:", value
                return value
        return
    
    ## Drawing Commands ##
    def _DRAW(self, args):
        return self.extend('_DRAW', args)
    
    def _DRAW_PXL_ON(self, args):
        self.emulator.rect(self.execute(args[0]), (1, 1))
        return 0
    
    def _DRAW_PXL_OFF(self, args):
        self.emulator.rect(self.execute(args[0]), (1, 1), color=(255,255,255))
        return 0
    
    def _DRAW_DISPGRAPH(self, args):
        self.emulator.update_screen()
        return 0
    
    def _DRAW_CLRDRAW(self, args):
        self.emulator.rect((0,0), self.emulator.size, color=(255,255,255))
        return 0
    
    def _DRAW_RECT(self, args):
        self.emulator.rect(self.execute(args[0]), self.execute(args[1]))
        return 0
    
    def _DRAW_RECTI(self, args):
        self.emulator.rect(self.execute(args[0]), self.execute(args[1]), color=(255,255,255))
        return 0
    
    def _DRAW_LINE(self, args):
        self.emulator.line(self.execute(args[0]), self.execute(args[1]))
        return 0
    
    def _COORD(self, args):
        return (self.execute(args[0]), self.execute(args[1]))
    
    
    ## Control Structures ##
    def _WHILE(self, args):
        while self.execute(args[0]):
            self.execute(args[1])
        return 0
    
    def _REPEAT(self, args):
        while not self.execute(args[0]):
            self.execute(args[1])
        return 0
            
    def _FOR(self, args):
        loop = args[0]
        var = loop[1]
        start = loop[2]
        end = loop[3]
        
        block = args[1]
        failsafe = 0
        
        i = start
        while i != end:
            self.execute(['ASSIGN', i, var])
            self.execute(block)
            i = self.execute(var)
            i += 1
        return -1   # Not sure why (but the original does this)
    
    def _IF(self, args):
        condition = self.execute(args[0])
        if condition:
            return self.execute(args[1])
        return 0
    
    def _CONDITION(self, args):
        return self.execute(args[0])
    
    def _OPERATOR(self, args):
        # TODO: implement overflow.
        op = args[0]
        value = self.execute(args[1])
        for a in args[2:]:
            value = int(getattr(operator, op)(value, self.execute(a)))
        value %= 65536
        return value
    
    def _ASSIGN(self, args):
        # TODO: Handle constants, not just vars.
        # TODO: implement overflow.
        # Note: This handles 2-byte vars
        
        value = self.execute(args[0]) % 65536   # 256 squared.
        self.vars[args[-1][-1]] = value
        return value
    
    def _BLOCK(self, args):
        new_args = []
        for arg in args:
            new_args.append(self.execute(arg))
        return new_args
    
    def _DISP(self, args):
        for arg in args:
            print 'OUT  :', self.execute(arg)
        return 0    # Disp doesn't return anything coherent?
    
    ### META ###
    def _META(self, args):
        self.extend('_META', args)
        return
    
    def _META_DEBUG(self, args):
        global _debug
        _debug = True if args[0] else False
        return args[0]
    
    def _META_STATE(self, args):
        print self.vars
        return 0
    
    def _META_DRAW(self, args):
        global _draw
        _draw = True if args[0] else False
        return args[0]
    
    def _META_HELP(self, args):
        print meta.help
        return 0
    
    def _META_ABOUT(self, args):
        print meta.about
        return 0
        
    def prune(self, ast):
        return ast
        name = ast[0]
        children = ast[1:]
        children_new = [name]
        for child in children:
            if type(child[1]) == list:
                if child[0] == 'IF' and child[1][1] == 0:
                    continue
                else:
                    new_child = self.prune(child)
                    children_new.append(new_child)
            else:
                children_new.append(child)
        return children_new

    def free(self, data):
        new_data = []
        for i in data:
            if type(i) == list:
                new_data.extend(self.free(i))
            else:
                new_data.append(i)
        return new_data
    
    def extend(self, head, args):
        return getattr(self, head + '_' + args[0])(args[1:])


def build():
    return Interpreter()


def test_interpreter(parser=axeparse.build(), lexer=axelex.build()):
    interpreter = Interpreter()
    global _draw
    _draw = True
    while 1:
        try:
            text = raw_input('calc > ')
        except (EOFError, KeyboardInterrupt):
            break
        if text[-1] != '\n' or text[-1] != ':':
            text = text + '\n'
        if text:
            try:
                result = parser.parse(text, lexer=lexer)
            except (ExitException):
                break
            if _debug:
                print result
                print interpreter.prune(result)
            print interpreter.run(result)
            print ''
    print '\nEnding...\n'

if __name__ == '__main__':
    test_interpreter()
    

