#!/usr/bin/env python
"""
Axe Interpreter

An interpreter for programs written in the Axe Parser language for computers.

Author: Michael Lee (<michael.lee.0x2a@gmail.com>)
License: GNU Lesser GPL
"""

## 3rd party libraries ##
from ply import lex as lex

## My Libraries ##
import io



## Define tokens ##

# Reserved keywords:
# Doesn't require a complex regex.
reserved = {
    'If' : 'IF',
    'End' : 'END',
    'Disp' : 'DISP',
    'For' : 'FOR',
    'While': 'WHILE',
    'Repeat': 'REPEAT',
    'Rect': 'RECT',
    'RectI': 'RECTI',
    'DispGraph': 'DISPGRAPH',
    'ClrDraw': 'CLRDRAW',
    'Line': 'LINE'
}

# Tokens:
# Requires a regex
tokens = [
    'DEBUG',
    'EXIT',
    'STATE',
    'DRAW',
    'HELP',
    'ABOUT',
    'NUMBER',
    'COLON',
    'COMMA',
    'NEWLINE',
    'LPAREN',
    'RPAREN',
    'ASSIGN',
    'PXL_ON',
    'PXL_OFF',
    'VAR',
    'ADD',
    'SUB',
    'MUL',
    'DIV',
    'LT',
    'LE',
    'EQ',
    'NE',
    'GE',
    'GT',
    'ID'
] + list(reserved.values())


## Define rules ##

# Note: the docstring is used for regex matching.


t_DEBUG = r'\@DEBUG'
t_EXIT = r'\@EXIT'
t_STATE = r'\@STATE'
t_DRAW = r'\@DRAW'
t_HELP = r'\@HELP'
t_ABOUT = r'\@ABOUT'

def t_NUMBER(t):
    r'[0-9]+'
    t.value = int(t.value)
    return t

def t_COMMENT(t):
    r'\..*'
    pass    # No return value.

def t_NEWLINE(t):
    r'(\n)+'
    t.lexer.lineno += len(t.value)
    t.value = 'NEWLINE'
    return t

def t_COLON(t):
    r'(:)+'
    t.value = 'NEWLINE'
    return t

def t_COMMA(t):
    r'(\,)'
    t.value = 'COMMA'
    return t

def t_LPAREN(t):
    r'(\()'
    t.value = 'LPAREN'
    return t

def t_RPAREN(t):
    r'(\))'
    t.value = 'RPAREN'
    return t

def t_ASSIGN(t):
    r'(\-\>)'
    t.value = 'ASSIGN'
    return t

def t_PXL_ON(t):
    r'Pxl\-On'
    t.value = 'PXL_ON'
    return t 

def t_PXL_OFF(t):
    r'Pxl\-Off'
    t.value = 'PXL_OFF'
    return t

def t_VAR(t):
    r'\b[A-Z]\b'
    t.value = 'VAR_' + t.value
    return t

def t_ADD(t):
    r'\+'
    t.value = 'ADD'
    return t
    
def t_SUB(t):
    r'\-'
    t.value = 'SUB'
    return t
    
def t_MUL(t):
    r'\*'
    t.value = 'MUL'
    return t
    
def t_DIV(t):
    r'\/'
    t.value = 'DIV'
    return t
    
def t_LT(t):
    r'\<'
    t.value = 'LT'
    return t
    
def t_LE(t):
    r'(\<\=){1}'
    t.value = 'LE'
    return t
    
def t_EQ(t):
    r'(\=)'
    t.value = 'EQ'
    return t
    
def t_NE(t):
    r'(\!\=)'
    t.value = 'NE'
    return t
    
def t_GE(t):
    r'(\>\=){1}'
    t.value = 'GE'
    return t
    
def t_GT(t):
    r'(\>){1}'
    t.value = 'GT'
    return t

def t_ID(t):
    r'[a-zA-Z_][a-zA-Z_0-9]*'
    t.type = reserved.get(t.value, 'ID')    # Check for reserved keywords
    return t

def t_error(t):
    error_str = "ERROR: Lexing.  Lineno {lineno}, column {column}.  Token: '{token}'"
    info = {'lineno': t.lexer.lineno,
            'column': find_column(t.value, t),
            'token':  t.value}
    print error_str.format(**info)
    t.lexer.skip(1)
    pass


t_ignore = ' \t'


## Misc. functions ##
def find_column(input_str, token):
    last_cr = input_str.rfind('\n',0,token.lexpos)
    if last_cr < 0:
        last_cr = 0
    column = (token.lexpos - last_cr) + 1
    return column
    
def build():
    return lex.lex(debug=0)


def test_lexer(text=''):
    lexer = build()
    if text != '':
        lexer.input(text)
        while 1:
            tok = lexer.token()
            if not tok:
                break
            print tok
    else:
        while 1:
            try:
                text = raw_input('calc > ')
            except EOFError:
                break
            except KeyboardInterrupt:
                print ''
                break
            if text:
                lexer.input(text)
                while 1:
                    tok = lexer.token()
                    if not tok:
                        break
                    print tok
    return

if __name__ == "__main__":
    test_lexer()







# Note: the below will slowly be added to the language.
# I'm not adding them all at once to avoid insanity.
"""
tokens = (
    "COLON",
    "PERIOD",
    "BUFFER",
    "CONSTANT",
    "ASSIGNMENT",
    "DEREFERENCE",
    "FUNCMODIFIER",
    "RMODIFIER",
    "RPAREN",
    "LPAREN"
    "RBRACKET",
    "LBRACKET",
    "RBRACE",
    "LBRACE",
    "NUMBER",
    "HEXADECIMAL",
    "STRING",
    "VARIABLE",
    "SINGLELINECOMMENT",
    "MULTILINECOMMENTSTART",
    "MULTILINECOMMENTEND",
    "LAMBDA",
)


reserved = {
    "DiagnosticOn":     "DIAGNOSTICON",
    "DiagnosticOff":    "DIAGNOSTICOFF",
    "Full":             "FULL",
    "Normal":           "NORMAL",
    "Pause":            "PAUSE",
    "GetKey":           "GETKEY",
    "Asm":              "ASM",
    "#Axiom":           "AXIOM",
    "#Icon":            "ICON",
    "#Realloc":         "REALLOC",
    "DispGraph":        "DISPGRAPH",
    "DispGraphClrDraw": "DISPGRAPHCLRDRAW",
    "ClrHome":          "CLRHOME",
    "ClrDraw":          "CLRDRAW",
    "DrawInv":          "DRAWINV",
    "StoreGDB":         "STOREGDB",
    "StorePic":         "STOREPIC",
    "RecallPic":        "RECALLPIC",
    "Horizontal":       "HORIZONTAL",
    "Vertical":         "VERTICAL",
    "Shade":            "SHADE",
    "If":               "IF",
    "Else":             "ELSE",
    "ElseIf":           "ELSEIF",
    "!If":              "NOTIF",
    "End":              "END",
    "Else!If":          "ELSENOTIF",
    "While":            "WHILE",
    "Repeat":           "REPEAT",
    "For":              "FOR",
    "DS<"               "DS",
    "EndIf":            "ENDIF",
    "End!If":           "ENDNOTIF",
    "Lbl":              "LBL",
    "Goto":             "GOTO",
    "Sub":              "SUB",
    "Return":           "RETURN",
    "ReturnIf":         "RETURNIF",
    "Return!If":        "RETURNNOTIF",
    "PLUS",
    "MINUS",
    "NEGATIVE",
    "INCREMENT",
    "DECREMENT",
    "TIMES",
    "DIVIDE",
    "MODULUS",
    "SQUARED",
    "EQUAL",
    "NOTEQUAL",
    "LESSTHEN",
    "GREATERTHEN",
    "LESSTHENOREQUAL",
    "GREATERTHENOREQUAL",
    "OR",
    "AND",
    "XOR",
    "ABS",
    "SQRT",
    "SIN",
    "COS",
    "EPOWER",
    "LN",
    "MIN",
    "MAX",
    "RAND",
    "SHORTIF",
    "SHORTTHEN",
    "SHORTNOTIF",
    "SHORTNOTTHEN",
    "HEXPREFIX",
    "BINARYPREFIX",
    "TOKENPREFIX",
    "DECIMAL",
    "LESSTHENLARGE",
    "LESSTHENEQUALLARGE",
    "GREATERTHENLARGE",
    "GREATERTHENEQUALLARGE",
    "SIGNEDMULTIPLY",
    "HIGHMULTIPLY",
    "SIGNEDDIVIDE"
    "ANDLARGE",
    "ORLARGE",
    "XORLARGE",
    "NOT",
    "EULER",
    "EULERLARGE",
    "RECIPROCAL",
    "PXLON",
    "PXLOFF",
    "PXLCHANGE",
    "PXLTEST",
    "LINE",
    "RECT",
    "RECTI",
    "CIRCLE",
    "PTON",
    "PTOFF",
    "PTCHANGE",
    "PTGET",
    "PTMASK",
    "BITMAP",
    "ROTC",
    "ROTCC",
    "FLIPV",
    "FLIPH",
    "DISP",
    "TODEC",
    "TOCHAR",
    "TOTOK",
    "TOXHEX",
    "HOMESCREENNEWLINE",
    "TEXT",
    "FIX",
    "PICVAR",
    "DATA",
    "BUFF",
    "SIGN",
    "NIB",
    "FILL",
    "COPY",
    "LENGTH",
    "INDATA",
    "STRCOMPARE",
    "SORTDATA",
    "CHECKSUM",
    "ANS",
    "GETCALC",
    "UNARCHIVE",
    "ARCHIVE",
    "DELVAR",
    "INPUT",
    "FLOAT",
    "FNINT",
    "FNON",
    "FNOFF",
    "STOP",
    "LNREG",
    "PORT",
    "FREQ",
    "SEND",
    "GET"
)


# Regex for simple tokens
t_PLUS = 
        "COLON",
    "PERIOD",
    "BUFFER",
    "CONSTANT",
    "ASSIGNMENT",
    "DEREFERENCE",
    "FUNCMODIFIER",
    "RMODIFIER",
    "RPAREN",
    "LPAREN"
    "RBRACKET",
    "LBRACKET",
    "RBRACE",
    "LBRACE",
    "NUMBER",
    "HEXADECIMAL",
    "STRING",
    "VARIABLE",
    "SINGLELINECOMMENT",
    "MULTILINECOMMENTSTART",
    "MULTILINECOMMENTEND",
    "DIAGNOSTICON",
    "DIAGNOSTICOFF",
    "FULL",
    "NORMAL",
    "PAUSE",
    "GETKEY",
    "ASM",
    "AXIOM",
    "ICON",
    "REALLOC",
    "DISPGRAPH",
    "DISPGRAPHCLRDRAW",
    "CLRHOME",
    "CLRDRAW",
    "DRAWINV",
    "STOREGDB",
    "STOREPIC",
    "RECALLPIC",
    "HORIZONTALPLUS",
    "HORIZONTALMINUS",
    "VERTICALPLUS",
    "VERTICALMINUS",
    "SHADE",
    "IF",
    "ELSE",
    "ELSEIF",
    "NOTIF",
    "END",
    "ELSENOTIF",
    "WHILE",
    "REPEAT",
    "FOR",
    "DS",
    "ENDIF",
    "ENDNOTIF",
    "LBL",
    "GOTO",
    "SUB",
    "RETURN",
    "RETURNIF",
    "RETURNNOTIF",
    "LAMBDA",
    "PLUS",
    "MINUS",
    "NEGATIVE",
    "INCREMENT",
    "DECREMENT",
    "TIMES",
    "DIVIDE",
    "MODULUS",
    "SQUARED",
    "EQUAL",
    "NOTEQUAL",
    "LESSTHEN",
    "GREATERTHEN",
    "LESSTHENOREQUAL",
    "GREATERTHENOREQUAL",
    "OR",
    "AND",
    "XOR",
    "ABS",
    "SQRT",
    "SIN",
    "COS",
    "EPOWER",
    "LN",
    "MIN",
    "MAX",
    "RAND",
    "SHORTIF",
    "SHORTTHEN",
    "SHORTNOTIF",
    "SHORTNOTTHEN",
    "HEXPREFIX",
    "BINARYPREFIX",
    "TOKENPREFIX",
    "DECIMAL",
    "LESSTHENLARGE",
    "LESSTHENEQUALLARGE",
    "GREATERTHENLARGE",
    "GREATERTHENEQUALLARGE",
    "SIGNEDMULTIPLY",
    "HIGHMULTIPLY",
    "SIGNEDDIVIDE"
    "ANDLARGE",
    "ORLARGE",
    "XORLARGE",
    "NOT",
    "EULER",
    "EULERLARGE",
    "RECIPROCAL",
    "PXLON",
    "PXLOFF",
    "PXLCHANGE",
    "PXLTEST",
    "LINE",
    "RECT",
    "RECTI",
    "CIRCLE",
    "PTON",
    "PTOFF",
    "PTCHANGE",
    "PTGET",
    "PTMASK",
    "BITMAP",
    "ROTC",
    "ROTCC",
    "FLIPV",
    "FLIPH",
    "DISP",
    "TODEC",
    "TOCHAR",
    "TOTOK",
    "TOXHEX",
    "HOMESCREENNEWLINE",
    "TEXT",
    "FIX",
    "PICVAR",
    "DATA",
    "BUFF",
    "SIGN",
    "NIB",
    "FILL",
    "COPY",
    "LENGTH",
    "INDATA",
    "STRCOMPARE",
    "SORTDATA",
    "CHECKSUM",
    "ANS",
    "GETCALC",
    "UNARCHIVE",
    "ARCHIVE",
    "DELVAR",
    "INPUT",
    "FLOAT",
    "FNINT",
    "FNON",
    "FNOFF",
    "STOP",
    "LNREG",
    "PORT",
    "FREQ",
    "SEND",
    "GET"
"""
    
