#!/usr/bin/env python
"""
Axe Interpreter

An interpreter for programs written in the Axe Parser language for computers.

Author: Michael Lee (<michael.lee.0x2a@gmail.com>)
License: GNU Lesser GPL
"""

# System libraries
import sys
import operator

# 3rd party libraries
from ply import lex as lex
from ply import yacc as yacc

# My libraries
import axelex
from axelex import tokens
from meta import ExitException

debug_ = False

def debug(*args):
    if debug_:
        print args
    return




## Program ##
def p_program(p):
    ''' program : program block
                | block'''
    debug('p_program',list(p))
    if len(p) == 3:
        p[0].append(p[2])
    elif len(p) == 2:
        p[0] = ['PROGRAM', p[1]]

def p_block_1(p):
    '''block : block block'''
    debug('p_statement_1', list(p))
    p[0] = ['BLOCK']
    p[0].extend(p[1][1:])
    p[0].extend(p[2][1:])
    return

def p_block_2(p):
    '''block : node
             | meta'''
    debug('p_block_2', list(p))
    p[0] = ['BLOCK', p[1]]
    return

def p_block_3(p):
    '''block : newline block'''
    debug('p_block_3', list(p))
    p[0] = p[2]
    return



## Nodes ##
def p_node_while(p):
    '''node : WHILE expression newline block END newline'''
    debug('p_block_while', list(p))
    p[0] = ['WHILE', ['CONDITION', p[2]], p[4]]
    return

def p_node_repeat(p):
    '''node : REPEAT expression newline block END newline'''
    debug('p_block_repeat', list(p))
    p[0] = ['REPEAT', ['CONDITION', p[2]], p[4]]
    return

def p_node_for(p):
    '''node : FOR LPAREN VAR COMMA expression COMMA expression RPAREN newline block END newline'''
    debug('p_block_for', list(p))
    
    var = p[3]
    start = p[5]
    end = p[7]
    block = p[10]
    p[0] = ['FOR', ['LOOP', var,start, end], block]
    return

def p_node_if(p):
    '''node : IF expression newline block END newline'''
    debug('p_block_if', list(p))
    p[0] = ['IF', ['CONDITION', p[2]], p[4]]
    return

def p_node_expr(p):
    '''node : expression newline'''
    debug('p_node_expr', list(p))
    p[0] = p[1]
    return

def p_node_disp(p):
    '''node : DISP expression newline'''
    debug('p_command_disp', list(p))
    p[0] = ['DISP', p[2]]
    return

def p_node_assign(p):
    '''node : assign newline'''
    debug('p_node_assign', list(p))
    p[0] = p[1]
    return

## Drawing Commands ##
def p_node_draw_pxl(p):
    '''node : PXL_ON LPAREN expression COMMA expression RPAREN newline
            | PXL_OFF LPAREN expression COMMA expression RPAREN newline'''
    debug('p_node_draw_pxl_on', list(p))
    p[0] = ['DRAW', p[1].upper(), ['COORD', p[3], p[5]]]
    return

def p_node_draw_dispgraph(p):
    '''node : DISPGRAPH newline'''
    debug('p_node_draw_dispgraph', list(p))
    p[0] = ['DRAW', p[1].upper()]
    return

def p_node_draw_clrdraw(p):
    '''node : CLRDRAW newline'''
    debug('p_node_draw_clrdraw', list(p))
    p[0] = ['DRAW', p[1].upper()]
    return

def p_node_draw_rect(p):
    '''node : RECT LPAREN expression COMMA expression COMMA expression COMMA expression RPAREN newline
            | RECTI LPAREN expression COMMA expression COMMA expression COMMA expression RPAREN newline'''
    debug('p_node_draw_rect', list(p))
    p[0] = ['DRAW', p[1].upper(), ['COORD', p[3], p[5]], ['COORD', p[7], p[9]]]
    return

def p_node_draw_line(p):
    '''node : LINE LPAREN expression COMMA expression COMMA expression COMMA expression RPAREN newline'''
    debug('p_node_draw_line', list(p))
    p[0] = ['DRAW', p[1].upper(), ['COORD', p[3], p[5]], ['COORD', p[7], p[9]]]
    return


## Assignment ##

def p_assign_seed(p):
    '''assign : expression ASSIGN VAR'''
    debug('p_node_assign', list(p))
    p[0] = ['ASSIGN', p[1],p[3]]
    return

def p_node_assign_chain(p):
    '''assign : assign ASSIGN VAR'''
    debug('p_node_assign_chain', list(p))
    p[0] = ['ASSIGN', p[1], p[3]]
    return

def p_node_assign_chain_2(p):
    '''assign : assign operator expression ASSIGN VAR'''
    debug('p_node_assign_chain_2', list(p))
    p[0] = ['ASSIGN', ['OPERATOR', p[2].lower(), p[3], p[1]], p[5]]
    
## Meta ##
def p_meta_exit(p):
    '''meta : EXIT newline
            | EXIT expression newline'''
    debug('p_meta_exit', list(p))
    exit_code = 'Called @EXIT'
    if len(p) == 4:
        exit_code = p[2]
    p[0] = ['META', 'EXIT', exit_code]
    
    raise ExitException(exit_code)
    return

def p_meta_debug(p):
    '''meta : DEBUG expression newline'''
    debug('p_meta_debug', list(p))
    global debug_
    if p[2] == 0:
        debug_ = False
    else:
        debug_ = True
    p[0] = ['META', 'DEBUG', p[2]]
    return

def p_meta_state(p):
    '''meta : STATE newline'''
    debug('p_meta_state', list(p))
    p[0] = ['META', 'STATE']
    return

def p_meta_draw(p):
    '''meta : DRAW expression newline'''
    debug('p_meta_draw', list(p))
    p[0] = ['META', 'DRAW', p[2]]
    return

def p_meta_help(p):
    '''meta : HELP newline'''
    debug('p_meta_help', list(p))
    p[0] = ['META', 'HELP']
    return

def p_meta_about(p):
    '''meta : ABOUT newline'''
    debug('p_meta_about', list(p))
    p[0] = ['META', 'ABOUT']
    return

## Numbers ##
def p_expression(p):
    ''' expression  : expression operator factor'''
    debug('p_expression', list(p))
    op = p[2].lower()
    if type(p[1]) == int and type(p[3]) == int:
        p[0] = int(getattr(operator, op)(p[1], p[3]))
    else:
        p[0] = ['OPERATOR', op, p[1], p[3]]
    return

def p_expression_num(p):
    '''expression : NUMBER'''
    debug('p_factor_num', list(p))
    p[0] = p[1]
    return

def p_expression_var(p):
    '''expression : VAR'''
    debug('p_factor_num', list(p))
    p[0] = p[1]
    return

def p_factor_paren(p):
    '''factor : LPAREN expression RPAREN'''
    debug('p_factor_paren', list(p))
    p[0] = p[2]
    return

def p_factor_num(p):
    '''factor : NUMBER
              | VAR'''
    debug('p_factor_num', list(p))
    p[0] = p[1]
    return

## Misc ##
def p_newline(p):
    '''newline : COLON
               | NEWLINE
               | newline newline''' # Reduces newlines.
    debug('p_newline', list(p))
    p[0] = p[1]
    return

def p_operator(p):
    '''operator : ADD
                | SUB
                | MUL
                | DIV
                | LT
                | LE
                | EQ
                | NE
                | GT
                | GE
    '''
    debug('p_operator', list(p))
    p[0] = p[1]
    return

def p_id(p):
    '''node : ID'''
    p[0] = None
    return

def p_error(t):
    try:
        info = {'lineno': t.lineno, 'value':t.value}
        print "Syntax error @ line {lineno}: '{value}'".format(**info)
    except AttributeError:
        print "Syntax error."
    return

def build():
    return yacc.yacc(tabmodule='parsing_rules', write_tables=1, debug=0)



def test_parser(lexer = axelex.build(), text=''):
    parser = build()
    if text != '':
        print parser.parse(text, lexer=lexer)
    else:
        while 1:
            try:
                s = raw_input('calc > ')
                if s[-1] != '\n' or s[-1] != ':':
                    s = s + '\n'
            except (EOFError, KeyboardInterrupt, ExitException):
                print ''
                break
            if s:
                result = parser.parse(s, lexer=lexer)
                print result, '\n'


if __name__ == "__main__":
    test_parser()

